﻿//---------------------------------------------------------------------------------------------
//	スペクトログラム
//		マイコン側のプログラム：TrG_Spectrogram
//
//		ポートの自動検出後のボー･レート：460800 baud
//		マイコンから受け取るデータ：１つのデータ当たり 0 ～ 10,000 に対応する2文字
//
//		ビルドの際にプロジェクトのプロパティで「アンセーフコードの許可」にチェックすること
//
//		作成者：三上直樹，2021/11/29 作成，©三上直樹 2021
//---------------------------------------------------------------------------------------------

using System;
using System.Drawing;
using System.Windows.Forms;
using System.IO.Ports;
using System.Drawing.Imaging;	// 高度な GDI+ イメージング機能

using Mikami.ComPort;
using Mikami.WindowsApplication;

namespace CQ_Spectrogram
{
	public partial class Form1 : Form
	{
		private const int BAUD_RATE_ = 460800;		// ボー･レート
		private const int N_FFT2_ = 256;			// FFT で使うデータ数の半分
		private const int SEC1_ = 10240/N_FFT2_;	// 1 秒に対応するピクセル数

		// シリアルポートと関連して使うクラスのオブジェクト
		private readonly MySerialPort sp_;
		// COM ポートの自動選択で使うクラスのオブジェクト
		private readonly AutoSelect comAuto_;
		private readonly int nY_;		// 縦（周波数方向）のピクセル数
		private readonly int nT_;		// 横（時間方向）のピクセル数

		private Color[,] twoD_;			// スペクトログラムに対応するデータ
		private Bitmap dst_;			// ビットマップ画像を扱うオブジェクト

		private float offset_ = 0;		// 表示の際のレベル調整で使う
		private bool run_ = false;		// RUN のとき true

		private string rxData_;			// 受信データ用バッファ
		private int nData_;				// データ数

		public Form1()
		{
			InitializeComponent();

			nY_ = pictureBox1.Size.Height;	// 縦（周波数方向）のピクセル数：251
			nT_ = pictureBox1.Size.Width;	// 横（時間方向）のピクセル数：480
			twoD_ = new Color[nT_, nY_];
			dst_= new Bitmap(nT_, nY_, PixelFormat.Format24bppRgb);

			CommonProc.Disable(button1, button2, button3);
			label4.Text = trackBar1.Value.ToString() + " dB";
			TwoD_Clear();
			run_ = true;

			sp_ = new MySerialPort(serialPort1);
			comAuto_ = new AutoSelect("Spectrogram", sp_, toolStripLabel1);
		}

		// マイコン側からのデータ取得実行の開始
		private void button1_Click(object sender, EventArgs e)
		{
			run_ = true;
			button1.Enabled = false;
			button2.Enabled = true;
			toolStripLabel1.Text = "RUN";
			RequestXfer();		// データ転送要求を送信
		}

		// 停止
		private void button2_Click(object sender, EventArgs e)
		{
			run_ = false;
			button1.Enabled = true;
			button2.Enabled = false;
			toolStripLabel1.Text = "STOP";
		}

		// クリア
		private void button3_Click(object sender, EventArgs e)
		{
			pictureBox1.Image = DrawBitmap(TwoD_Clear());
		}

		// フォームを閉じることの確認と，閉じてもよいときに
		// シリアルポートがオープンであれば，これをクローズする
		private void Form1_FormClosing(object sender, FormClosingEventArgs e)
		{
			bool runSave = run_;
			run_ = false;
			DialogResult result =
				MessageBox.Show("閉じてもいいですか？",
								"終了の確認",
								MessageBoxButtons.YesNo,
								MessageBoxIcon.Information);
			if (result == DialogResult.No)
			{
				e.Cancel = true;	// 閉じるのをキャンセルする
				if (runSave)
				{
					run_ = true;
					RequestXfer();	// データ転送要求を送信
				}
				return;
			}

			if (!serialPort1.IsOpen) return;

			try {
				serialPort1.BreakState = true;	// マイコンのリセット
				serialPort1.BreakState = false;
				serialPort1.Close();			// シリアルポートのクローズ
			}
			catch (Exception ex) { sp_.ExceptionCatch(ex); }
		}

		// 座標軸等の描画
		private void Form1_Paint(object sender, PaintEventArgs e)
		{
			DrawAxis(e.Graphics);
		}

		// レベル調整用トラックバーの値の表示
		private void trackBar1_Scroll(object sender, EventArgs e)
		{
			offset_ = trackBar1.Value;
			label4.Text = offset_.ToString() + " dB";
		}

		// データ受信イベントに対応するハンドラ
		private void serialPort1_DataReceived(object sender, SerialDataReceivedEventArgs e)
		{
			try {
				rxData_ = serialPort1.ReadLine();
				if (rxData_ == "ACK")
				{
					comAuto_.StopTimer();	// タイムアウト監視タイマを停止
					toolStripLabel1.Text = "マイコン側の準備ができました．";
					toolStripLabel2.Text = " " + serialPort1.PortName + " ";
					serialPort1.BaudRate = BAUD_RATE_;	// ボー･レートを高速に設定する
					Invoke(() => CommonProc.Enable(button1, button3));
				}
				else
					DrawSpectrogram();		// "ACK" 受信以外の処理
			}
			catch (Exception ex) { sp_.ExceptionCatch(ex); }
		}
	}
}
