﻿//--------------------------------------------------------------
//	スペクトログラムで使うメソッド
//
//		作成者：三上直樹，2021/11/29 作成，©三上直樹 2021
//--------------------------------------------------------------

using System;
using System.Drawing;
using System.Windows.Forms;
using System.Drawing.Imaging;	// 高度な GDI+ イメージング機能

using static System.Math;

namespace CQ_Spectrogram
{
	public partial class Form1 : Form
	{
		// 座標軸などの描画
		private void DrawAxis(Graphics gr)
		{
			int orgX = pictureBox1.Location.X;
			int orgY = pictureBox1.Location.Y + nY_ - 1;
			gr.TranslateTransform(orgX, orgY);	// 座標軸の原点の移動
			Font myFont = new Font("Meiryo UI", 9);
			Brush myBrush = Brushes.Black;

			// 座標軸の描画（横）
			gr.DrawLine(Pens.Black, 0, 5, nT_, 5);

			for (int n=0; n<=(int)(nT_/SEC1_); n++)		// 目盛
				gr.DrawLine(Pens.Black, n*SEC1_, 1, n*SEC1_, 5);
			for (int n=0; n<=(int)(nT_/SEC1_); n+=2)	// 目盛値
				if (n < 10)
					gr.DrawString(n.ToString(), myFont, myBrush, n*SEC1_-6, 9);
				else
					gr.DrawString(n.ToString(), myFont, myBrush, n*SEC1_-10, 9);
			gr.DrawString("時間 [s]", myFont, myBrush, nT_/2-30, 26);

			// 座標軸の描画（縦）
			gr.DrawLine(Pens.Black, -5, 0, -5, -nY_);
			const int MAX_FRQ = 5;		// 表示する最大の周波数（kHz）
			float dy = nY_/(float)MAX_FRQ;
			for (int n=0; n<=MAX_FRQ; n++)
				gr.DrawLine(Pens.Black, -5, -n*dy, -1, -n*dy);
			for (int n=0; n<=MAX_FRQ; n++)
				gr.DrawString(n.ToString(), myFont, myBrush, -20, -n*dy-8);

			gr.RotateTransform(-90);
			gr.DrawString("周波数 [kHz]", myFont, myBrush, 86, -48);
			gr.ResetTransform();

			// 色と dB の関係
			gr.TranslateTransform(pictureBox1.Location.X + pictureBox1.Width + 46,
								  pictureBox1.Location.Y + pictureBox1.Size.Height);
			const int Y20DB = (int)(2.5f*20);	// 20 dB に対応するピクセル数
			ColorDb(gr, myFont, myBrush, Y20DB);

			int x1 = 4;
			int x2 = x1 + 10;
			float yF = 1.0f/(Y20DB*4);
			Pen myPen = new Pen(Color.Transparent, 1);
			for (int n = 0; n <= Y20DB*4; n++)
			{
				myPen.Color = HueScale(n*yF);
				gr.DrawLine(myPen, x1, -n, x2, -n);
			}
		}

		// スペクトログラムの描画
		void DrawSpectrogram()
		{
			nData_ = rxData_.Length/2;

			// 表示用バッファのデータを一つ左に移動
			for (int j=1; j<nT_; j++)
				for (int k=0; k<nY_; k++)
					twoD_[j-1, k] = twoD_[j, k];

			// 表示用バッファの更新
			const float FACTOR = 1.0f/80.0f;
			for (int n=0; n<nData_; n++)
			{
				float x = Decode(rxData_.Substring(2*n, 2));
				float db = FACTOR*(float)(20.0*Log10(x) + 76.0 + offset_);
				if (db < 0) db = 0;
				twoD_[nT_-1, nData_-n-1] = HueScale(db);
			}

			Invoke(() => pictureBox1.Image = DrawBitmap(twoD_) );
			if (run_) RequestXfer();	// データ転送要求を送信
		}

		// 色と dB の関係の座標軸の描画（縦）
		private void ColorDb(Graphics gr, Font myFont, Brush myBrush, int y20dB)
		{
			gr.DrawLine(Pens.Black, -5, 0, -5, -y20dB*4);
			for (int n = 0; n <= 8; n++)
				gr.DrawLine(Pens.Black, -5, -n*y20dB/2, -1, -n*y20dB/2);
			gr.DrawString("0", myFont, myBrush, -22, -8);
			for (int n = 1; n <= 4; n++)
				gr.DrawString((n*20).ToString(), myFont, myBrush, -30, -n*y20dB-8);
			gr.DrawString("[dB]", myFont, myBrush, -36, -4*y20dB-24);
		}

		// 表示用バッファのクリア
		private Color[,] TwoD_Clear()
		{
			for (int j=0; j<nT_; j++)
				for (int k=0; k<nY_; k++)
					twoD_[j, k] = SystemColors.ControlLightLight;
			return twoD_;
		}

		// 色相の違いで表示
		//		0.0 <= x <= 1.0
		//		x > 1 の場合　戻り値： Color.White
		//		x < 0 の場合　戻り値： Color.Black
		private Color HueScale(float x)
		{
			if (x > 1.0f) return Color.White;
			if (x < 0.0f) return Color.Black;

			int r = (x >= 0.5f) ? 255 : 0;
			if ((x > 0.25f) && (x < 0.5)) r = -255 + (int)(1020*x);
			int b = (x < 0.5f) ? 255 - (int)(510*x) : 0;
			int g = -(int)(1020*x*(x - 1));

			return Color.FromArgb(r, g, b);
		}

		// ビットマップへの高速転送
		//		1 ピクセル当たり 24 ビット割り当て
		Bitmap DrawBitmap(Color[,] twoD)
		{
			// Bitmap のメモリをロックする
			BitmapData dstData = dst_.LockBits(new Rectangle(Point.Empty, dst_.Size),
											   ImageLockMode.WriteOnly,
											   dst_.PixelFormat);
			int myStride = dstData.Stride;

			// プロジェクトのビルドに関するプロパティを「アンセーフコードの許可」にすること
			unsafe
			{
				byte *dstScan0 = (byte *)dstData.Scan0;	// ピクセルデータのポインタを取得

				for (int jy = 0; jy < nY_; jy++)
					for (int jx = 0; jx < nT_; jx++)
					{
						int dPos = 3*jx + myStride*jy;
						dstScan0[dPos+2] = twoD[jx, jy].R;	// R
						dstScan0[dPos+1] = twoD[jx, jy].G;	// G
						dstScan0[dPos+0] = twoD[jx, jy].B;	// B
					}
			}
			dst_.UnlockBits(dstData);

			return dst_;
		}

		// 0 ～ 10,000 の範囲の数を 2 文字で表したコードを float 型に変換する
		// 10,000 を 1.0f に対応させる
		private float Decode(string s)
		{
			uint x = (((uint)s[0] - 0x10) << 7) | ((uint)s[1] - 0x10);
			#if DEBUG
				if (x > 10000) MessageBox.Show(x.ToString());
			#endif
			return x/10000.0f;
		}

		// データの転送を要求する
		private void RequestXfer()
		{
			// 転送要求と高域強調の有無
			string str = "GO" + (radioButton1.Checked ? "Y" : "N");
			sp_.SendTo(str);
		}

		// 別のスレッドを使う場合に使用
		private void Invoke(Action action) => base.Invoke(action);
	}
}
